@@ -2,6 +2,7 @@
 #define _POSIX_C_SOURCE 199309L
 #include <iostream>
 
+#include <list>
 #include <string>
 #include <thread>
 #include <mutex>
@@ -13,27 +14,31 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/mman.h>
+#include <sys/stat.h>
 #include <signal.h>
 #include <unistd.h>
 #include <wayland-client-protocol.h>
 
 #include "frame-writer.hpp"
-#include "pulse.hpp"
 #include "wlr-screencopy-unstable-v1-client-protocol.h"
 #include "xdg-output-unstable-v1-client-protocol.h"
 
 #include "config.h"
 
-#ifdef HAVE_OPENCL
-std::unique_ptr<OpenCL> opencl;
+#ifdef HAVE_PULSE
+#include "pulse.hpp"
+PulseReaderParams pulseParams;
 #endif
 
+#define MAX_FRAME_FAILURES 16
+
 std::mutex frame_writer_mutex, frame_writer_pending_mutex;
 std::unique_ptr<FrameWriter> frame_writer;
 
 static struct wl_shm *shm = NULL;
 static struct zxdg_output_manager_v1 *xdg_output_manager = NULL;
 static struct zwlr_screencopy_manager_v1 *screencopy_manager = NULL;
+void request_next_frame();
 
 struct wf_recorder_output
 {
@@ -43,7 +48,7 @@ struct wf_recorder_output
     int32_t x, y, width, height;
 };
 
-std::vector<wf_recorder_output> available_outputs;
+std::list<wf_recorder_output> available_outputs;
 
 static void handle_xdg_output_logical_position(void*,
     zxdg_output_v1* zxdg_output, int32_t x, int32_t y)
@@ -174,6 +179,8 @@ static struct wl_buffer *create_shm_buffer(uint32_t fmt,
     return buffer;
 }
 
+static bool use_damage = true;
+
 static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame, uint32_t format,
     uint32_t width, uint32_t height, uint32_t stride)
 {
@@ -184,6 +191,12 @@ static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame,
     buffer.height = height;
     buffer.stride = stride;
 
+    /* ffmpeg requires even width and height */
+    if (buffer.width % 2)
+        buffer.width -= 1;
+    if (buffer.height % 2)
+        buffer.height -= 1;
+
     if (!buffer.wl_buffer) {
         buffer.wl_buffer =
             create_shm_buffer(format, width, height, stride, &buffer.data);
@@ -194,13 +207,19 @@ static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame,
         exit(EXIT_FAILURE);
     }
 
-    zwlr_screencopy_frame_v1_copy(frame, buffer.wl_buffer);
+    if (use_damage) {
+        zwlr_screencopy_frame_v1_copy_with_damage(frame, buffer.wl_buffer);
+    } else {
+        zwlr_screencopy_frame_v1_copy(frame, buffer.wl_buffer);
+    }
 }
 
 static void frame_handle_flags(void*, struct zwlr_screencopy_frame_v1 *, uint32_t flags) {
     buffers[active_buffer].y_invert = flags & ZWLR_SCREENCOPY_FRAME_V1_FLAGS_Y_INVERT;
 }
 
+int32_t frame_failed_cnt = 0;
+
 static void frame_handle_ready(void *, struct zwlr_screencopy_frame_v1 *,
     uint32_t tv_sec_hi, uint32_t tv_sec_low, uint32_t tv_nsec) {
 
@@ -208,11 +227,22 @@ static void frame_handle_ready(void *, struct zwlr_screencopy_frame_v1 *,
     buffer_copy_done = true;
     buffer.presented.tv_sec = ((1ll * tv_sec_hi) << 32ll) | tv_sec_low;
     buffer.presented.tv_nsec = tv_nsec;
+    frame_failed_cnt = 0;
 }
 
 static void frame_handle_failed(void *, struct zwlr_screencopy_frame_v1 *) {
-    fprintf(stderr, "failed to copy frame\n");
-    exit_main_loop = true;
+    std::cerr << "Failed to copy frame, retrying..." << std::endl;
+    ++frame_failed_cnt;
+    if (frame_failed_cnt > MAX_FRAME_FAILURES)
+    {
+        std::cerr << "Failed to copy frame too many times, exiting!" << std::endl;
+        exit_main_loop = true;
+    }
+}
+
+static void frame_handle_damage(void *, struct zwlr_screencopy_frame_v1 *,
+    uint32_t, uint32_t, uint32_t, uint32_t)
+{
 }
 
 static const struct zwlr_screencopy_frame_v1_listener frame_listener = {
@@ -220,6 +250,7 @@ static const struct zwlr_screencopy_frame_v1_listener frame_listener = {
     .flags = frame_handle_flags,
     .ready = frame_handle_ready,
     .failed = frame_handle_failed,
+    .damage = frame_handle_damage,
 };
 
 static void handle_global(void*, struct wl_registry *registry,
@@ -239,7 +270,7 @@ static void handle_global(void*, struct wl_registry *registry,
     else if (strcmp(interface, zwlr_screencopy_manager_v1_interface.name) == 0)
     {
         screencopy_manager = (zwlr_screencopy_manager_v1*) wl_registry_bind(registry, name,
-            &zwlr_screencopy_manager_v1_interface, 1);
+            &zwlr_screencopy_manager_v1_interface, 2);
     }
     else if (strcmp(interface, zxdg_output_manager_v1_interface.name) == 0)
     {
@@ -269,21 +300,32 @@ static int next_frame(int frame)
 
 static InputFormat get_input_format(wf_buffer& buffer)
 {
-    if (buffer.format == WL_SHM_FORMAT_ARGB8888)
+    switch (buffer.format) {
+    case WL_SHM_FORMAT_ARGB8888:
+    case WL_SHM_FORMAT_XRGB8888:
         return INPUT_FORMAT_BGR0;
-    if (buffer.format == WL_SHM_FORMAT_XRGB8888)
-        return INPUT_FORMAT_BGR0;
-
-    if (buffer.format == WL_SHM_FORMAT_XBGR8888)
+    case WL_SHM_FORMAT_XBGR8888:
+    case WL_SHM_FORMAT_ABGR8888:
         return INPUT_FORMAT_RGB0;
-    if (buffer.format == WL_SHM_FORMAT_ABGR8888)
-        return INPUT_FORMAT_RGB0;
-
-    fprintf(stderr, "Unsupported buffer format %d, exiting.", buffer.format);
-    std::exit(0);
+    case WL_SHM_FORMAT_BGR888:
+        return INPUT_FORMAT_BGR8;
+    case WL_SHM_FORMAT_RGB565:
+        return INPUT_FORMAT_RGB565;
+    case WL_SHM_FORMAT_BGR565:
+        return INPUT_FORMAT_BGR565;
+    case WL_SHM_FORMAT_ARGB2101010:
+    case WL_SHM_FORMAT_XRGB2101010:
+        return INPUT_FORMAT_X2RGB10;
+    case WL_SHM_FORMAT_ABGR2101010:
+    case WL_SHM_FORMAT_XBGR2101010:
+        return INPUT_FORMAT_X2BGR10;
+    default:
+        fprintf(stderr, "Unsupported buffer format %d, exiting.", buffer.format);
+        std::exit(0);
+    }
 }
 
-static void write_loop(FrameWriterParams params, PulseReaderParams pulseParams)
+static void write_loop(FrameWriterParams params)
 {
     /* Ignore SIGINT, main loop is responsible for the exit_main_loop signal */
     sigset_t sigset;
@@ -292,12 +334,14 @@ static void write_loop(FrameWriterParams params, PulseReaderParams pulseParams)
     pthread_sigmask(SIG_BLOCK, &sigset, NULL);
 
     int last_encoded_frame = 0;
+#ifdef HAVE_PULSE
     std::unique_ptr<PulseReader> pr;
+#endif
 
     while(!exit_main_loop)
     {
         // wait for frame to become available
-        while(buffers[last_encoded_frame].available != true) {
+        while(buffers[last_encoded_frame].available != true && !exit_main_loop) {
             std::this_thread::sleep_for(std::chrono::microseconds(1000));
         }
         auto& buffer = buffers[last_encoded_frame];
@@ -312,26 +356,24 @@ static void write_loop(FrameWriterParams params, PulseReaderParams pulseParams)
             params.format = get_input_format(buffer);
             params.width = buffer.width;
             params.height = buffer.height;
+            params.stride = buffer.stride;
             frame_writer = std::unique_ptr<FrameWriter> (new FrameWriter(params));
 
-#ifdef HAVE_OPENCL
-            if (params.opencl && params.force_yuv)
-            {
-                frame_writer->opencl = std::move(opencl);
-                frame_writer->opencl->init(params.width, params.height);
-            }
-#endif
-
+#ifdef HAVE_PULSE
             if (params.enable_audio)
             {
                 pulseParams.audio_frame_size = frame_writer->get_audio_buffer_size();
                 pr = std::unique_ptr<PulseReader> (new PulseReader(pulseParams));
                 pr->start();
             }
+#endif
         }
 
-        frame_writer->add_frame((unsigned char*)buffer.data, buffer.base_usec,
-            buffer.y_invert);
+        bool do_cont = frame_writer->add_frame((unsigned char*)buffer.data,
+            buffer.base_usec, buffer.y_invert);
+        if (!do_cont) {
+            break;
+        }
 
         frame_writer_mutex.unlock();
 
@@ -344,7 +386,9 @@ static void write_loop(FrameWriterParams params, PulseReaderParams pulseParams)
     std::lock_guard<std::mutex> lock(frame_writer_mutex);
     /* Free the PulseReader connection first. This way it'd flush any remaining
      * frames to the FrameWriter */
+#ifdef HAVE_PULSE
     pr = nullptr;
+#endif
     frame_writer = nullptr;
 }
 
@@ -353,6 +397,24 @@ void handle_sigint(int)
     exit_main_loop = true;
 }
 
+static bool user_specified_overwrite(std::string filename)
+{
+    struct stat buffer;   
+    if (stat (filename.c_str(), &buffer) == 0)
+    {
+        std::string input;
+        std::cout << "Output file \"" << filename << "\" exists. Overwrite? Y/n: ";
+        std::getline(std::cin, input);
+        if (input.size() && input[0] != 'Y' && input[0] != 'y')
+        {
+            std::cout << "Use -f to specify the file name." << std::endl;
+            return false;
+	}
+    }
+
+    return true;
+}
+
 static void check_has_protos()
 {
     if (shm == NULL) {
@@ -416,7 +478,9 @@ static wf_recorder_output* choose_interactive()
     if (scanf("%d", &choice) != 1 || choice > (int)available_outputs.size() || choice <= 0)
         return nullptr;
 
-    return &available_outputs[choice - 1];
+    auto it = available_outputs.begin();
+    std::advance(it, choice - 1);
+    return &*it;
 }
 
 struct capture_region
@@ -430,22 +494,6 @@ struct capture_region
     capture_region(int32_t _x, int32_t _y, int32_t _width, int32_t _height)
         : x(_x), y(_y), width(_width), height(_height) { }
 
-    /* Make sure that dimension is even, while trying to keep the segment
-     * [coordinate, coordinate+dimension) as good as possible (i.e not going
-     * out of the monitor) */
-    void make_even(int32_t& coordinate, int32_t& dimension)
-    {
-        if (dimension % 2 == 0)
-            return;
-
-        /* We need to increase dimension to make it an even number */
-        ++dimension;
-
-        /* Try to decrease coordinate. If coordinate > 0, we can always lower it
-         * by 1 pixel and stay inside the screen. */
-        coordinate = std::max(coordinate - 1, 0);
-    }
-
     void set_from_string(std::string geometry_string)
     {
         if (sscanf(geometry_string.c_str(), "%d,%d %dx%d", &x, &y, &width, &height) != 4)
@@ -455,11 +503,6 @@ struct capture_region
             x = y = width = height = 0;
             return;
         }
-
-        /* ffmpeg requires even width and height */
-        make_even(x, width);
-        make_even(y, height);
-        printf("Adjusted geometry: %d,%d %dx%d\n", x, y, width, height);
     }
 
     bool is_selected()
@@ -500,11 +543,17 @@ Screen recording of wlroots-based compositors
 
 With no FILE, start recording the current screen.
 
+Use Ctrl+C to stop.)");
+#ifdef HAVE_PULSE
+    printf(R"(
+
   -a, --audio [DEVICE]      Starts recording the screen with audio.
                             [DEVICE] argument is optional.
                             In case you want to specify the pulseaudio device which will capture
                             the audio, you can run this command with the name of that device.
-                            You can find your device by running: pactl list sinks | grep Name
+                            You can find your device by running: pactl list sources | grep Name)");
+#endif
+    printf(R"(
 
   -c, --codec               Specifies the codec of the video. Supports  GIF output also.
                             To modify codec parameters, use -p <option_name>=<option_value>
@@ -513,6 +562,12 @@ With no FILE, start recording the current screen.
                             Some drivers report support for rgb0 data for vaapi input but
                             really only support yuv.
 
+  -D, --no-damage           By default, wf-recorder will request a new frame from the compositor
+                            only when the screen updates. This results in a much smaller output
+                            file, which however has a variable refresh rate. When this option is
+                            on, wf-recorder does not use this optimization and continuously
+                            records new frames, even if there are no updates on the screen.
+
   -f <filename>.ext         By using the -f option the output file will have the name :
                             filename.ext and the file format will be determined by provided
                             while extension .ext . If the extension .ext provided is not
@@ -526,33 +581,34 @@ With no FILE, start recording the current screen.
   -x, --pixel-format        Set the output pixel format. These can be found by running:
                             *ffmpeg -pix_fmts*
 
-  -g, --geometry            Selects a specific part of the screen.
+  -g, --geometry            Selects a specific part of the screen. The format is "x,y WxH".
 
   -h, --help                Prints this help screen.
 
+  -v, --version             Prints the version of wf-recorder.
+
   -l, --log                 Generates a log on the current terminal. Debug purposes.
 
   -o, --output              Specify the output where the video is to be recorded.
 
   -p, --codec-param         Change the codec parameters.
-                            -p <option_name>=<option_value>)");
-#ifdef HAVE_OPENCL
-    printf(R"(
+                            -p <option_name>=<option_value>
 
-  -e, --opencl              Use the -e[#] or --opencl[=#] in conjunction with -t or --force-yuv option
-                            to use opencl for gpu accelerated conversion of data to yuv. # is one
-                            of the devices listed when running without specifying #.)");
-#endif
-    printf(R"(
+  -F, --filter              Specify the ffmpeg filter string to use. For example,
+                            -F hwupload,scale_vaapi=format=nv12 is used for VAAPI.
 
   -t, --force-yuv           Use the -t or --force-yuv option to force conversion of the data to
                             yuv format, before sending it to the gpu.
 
   -b, --bframes             This option is used to set the maximum number of b-frames to be used.
                             If b-frames are not supported by your hardware, set this to 0.)" "\n\n" R"(
-Examples:
+Examples:)");
+#ifdef HAVE_PULSE
+    printf(R"(
 
-  Video Only:
+  Video Only:)");
+#endif
+    printf(R"(
 
   - wf-recorder                         Records the video. Use Ctrl+C to stop recording.
                                         The video file will be stored as recording.mp4 in the
@@ -560,7 +616,9 @@ Examples:
 
   - wf-recorder -f <filename>.ext       Records the video. Use Ctrl+C to stop recording.
                                         The video file will be stored as <outputname>.ext in the
-                                        current working directory.
+                                        current working directory.)");
+#ifdef HAVE_PULSE
+    printf(R"(
 
   Video and Audio:
 
@@ -570,31 +628,56 @@ Examples:
 
   - wf-recorder -a -f <filename>.ext    Records the audio. Use Ctrl+C to stop recording.
                                         The video file will be stored as <outputname>.ext in the
-                                        current working directory.
+                                        current working directory.)");
+#endif
+    printf(R"(
 
 )" "\n");
     exit(EXIT_SUCCESS);
 }
 
+capture_region selected_region{};
+wf_recorder_output *chosen_output = nullptr;
+zwlr_screencopy_frame_v1 *frame = NULL;
+
+void request_next_frame()
+{
+    if (frame != NULL)
+    {
+        zwlr_screencopy_frame_v1_destroy(frame);
+    }
+
+    /* Capture the whole output if the user hasn't provided a good geometry */
+    if (!selected_region.is_selected())
+    {
+        frame = zwlr_screencopy_manager_v1_capture_output(
+            screencopy_manager, 1, chosen_output->output);
+    } else
+    {
+        frame = zwlr_screencopy_manager_v1_capture_output_region(
+            screencopy_manager, 1, chosen_output->output,
+            selected_region.x - chosen_output->x,
+            selected_region.y - chosen_output->y,
+            selected_region.width, selected_region.height);
+    }
+
+    zwlr_screencopy_frame_v1_add_listener(frame, &frame_listener, NULL);
+}
+
+
 int main(int argc, char *argv[])
 {
-    FrameWriterParams params;
+    FrameWriterParams params = FrameWriterParams(exit_main_loop);
     params.file = "recording.mp4";
     params.codec = DEFAULT_CODEC;
     params.enable_ffmpeg_debug_output = false;
     params.enable_audio = false;
     params.force_yuv = false;
-    params.opencl = false;
-    params.opencl_device = -1;
     params.bframes = -1;
 
-    PulseReaderParams pulseParams;
-
     constexpr const char* default_cmdline_output = "interactive";
     std::string cmdline_output = default_cmdline_output;
 
-    capture_region selected_region{};
-
     struct option opts[] = {
         { "output",          required_argument, NULL, 'o' },
         { "file",            required_argument, NULL, 'f' },
@@ -604,19 +687,21 @@ int main(int argc, char *argv[])
         { "codec",           required_argument, NULL, 'c' },
         { "codec-param",     required_argument, NULL, 'p' },
         { "device",          required_argument, NULL, 'd' },
+        { "filter",          required_argument, NULL, 'F' },
         { "log",             no_argument,       NULL, 'l' },
         { "audio",           optional_argument, NULL, 'a' },
         { "help",            no_argument,       NULL, 'h' },
         { "force-yuv",       no_argument,       NULL, 't' },
-        { "opencl",          optional_argument, NULL, 'e' },
         { "bframes",         required_argument, NULL, 'b' },
+        { "version",         no_argument,       NULL, 'v' },
+        { "no-damage",       no_argument,       NULL, 'D' },
         { 0,                 0,                 NULL,  0  }
     };
 
     int c, i;
     std::string param;
     size_t pos;
-    while((c = getopt_long(argc, argv, "o:f:m:x:g:c:p:d:b:la::te::h", opts, &i)) != -1)
+    while((c = getopt_long(argc, argv, "o:f:m:x:g:c:p:d:b:la::thvDF:", opts, &i)) != -1)
     {
         switch(c)
         {
@@ -624,6 +709,10 @@ int main(int argc, char *argv[])
                 params.file = optarg;
                 break;
 
+            case 'F':
+                params.video_filter = optarg;
+                break;
+
             case 'o':
                 cmdline_output = optarg;
                 break;
@@ -657,8 +746,13 @@ int main(int argc, char *argv[])
                 break;
 
             case 'a':
+#ifdef HAVE_PULSE
                 params.enable_audio = true;
                 pulseParams.audio_source = optarg ? strdup(optarg) : NULL;
+#else
+                std::cerr << "Cannot record audio. Built without pulse support." << std::endl;
+                return EXIT_FAILURE;
+#endif
                 break;
 
             case 't':
@@ -669,11 +763,6 @@ int main(int argc, char *argv[])
                 help();
                 break;
 
-            case 'e':
-                params.opencl = true;
-                params.opencl_device = optarg ? atoi(optarg) : -1;
-                break;
-
             case 'p':
                 param = optarg;
                 pos = param.find("=");
@@ -688,13 +777,26 @@ int main(int argc, char *argv[])
                 }
                 break;
 
+            case 'v':
+                printf("wf-recorder %s\n", WFRECORDER_VERSION);
+                return 0;
+
+            case 'D':
+                use_damage = false;
+                break;
+
             default:
                 printf("Unsupported command line argument %s\n", optarg);
         }
     }
 
+    if (!user_specified_overwrite(params.file))
+    {
+        return EXIT_FAILURE;
+    }
+
     display = wl_display_connect(NULL);
-    if (display == NULL) 
+    if (display == NULL)
     {
         fprintf(stderr, "failed to create display: %m\n");
         return EXIT_FAILURE;
@@ -707,10 +809,9 @@ int main(int argc, char *argv[])
     check_has_protos();
     load_output_info();
 
-    wf_recorder_output *chosen_output = nullptr;
     if (available_outputs.size() == 1)
     {
-        chosen_output = &available_outputs[0];
+        chosen_output = &available_outputs.front();
         if (chosen_output->name != cmdline_output &&
             cmdline_output != default_cmdline_output)
         {
@@ -764,12 +865,7 @@ int main(int argc, char *argv[])
         }
     }
 
-    printf("selected region %d %d %d %d\n", selected_region.x, selected_region.y, selected_region.width, selected_region.height);
-
-#ifdef HAVE_OPENCL
-     if (params.opencl && params.force_yuv)
-         opencl = std::unique_ptr<OpenCL> (new OpenCL(params.opencl_device));
-#endif
+    printf("selected region %d,%d %dx%d\n", selected_region.x, selected_region.y, selected_region.width, selected_region.height);
 
     timespec first_frame;
     first_frame.tv_sec = -1;
@@ -795,26 +891,14 @@ int main(int argc, char *argv[])
         }
 
         buffer_copy_done = false;
-        struct zwlr_screencopy_frame_v1 *frame = NULL;
+        request_next_frame();
 
-        /* Capture the whole output if the user hasn't provided a good geometry */
-        if (!selected_region.is_selected())
-        {
-            frame = zwlr_screencopy_manager_v1_capture_output(
-                screencopy_manager, 1, chosen_output->output);
-        } else
-        {
-            frame = zwlr_screencopy_manager_v1_capture_output_region(
-                screencopy_manager, 1, chosen_output->output,
-                selected_region.x - chosen_output->x,
-                selected_region.y - chosen_output->y,
-                selected_region.width, selected_region.height);
+        while (!buffer_copy_done && !exit_main_loop && wl_display_dispatch(display) != -1) {
+            // This space is intentionally left blank
         }
 
-        zwlr_screencopy_frame_v1_add_listener(frame, &frame_listener, NULL);
-
-        while (!buffer_copy_done && wl_display_dispatch(display) != -1) {
-            // This space is intentionally left blank
+        if (exit_main_loop) {
+            break;
         }
 
         auto& buffer = buffers[active_buffer];
@@ -823,7 +907,7 @@ int main(int argc, char *argv[])
         if (!spawned_thread)
         {
             writer_thread = std::thread([=] () {
-                write_loop(params, pulseParams);
+                write_loop(params);
             });
 
             spawned_thread = true;
@@ -839,13 +923,15 @@ int main(int argc, char *argv[])
         buffer.available = true;
 
         active_buffer = next_frame(active_buffer);
-        zwlr_screencopy_frame_v1_destroy(frame);
     }
 
     writer_thread.join();
 
     for (auto& buffer : buffers)
-        wl_buffer_destroy(buffer.wl_buffer);
+    {
+        if (buffer.wl_buffer)
+            wl_buffer_destroy(buffer.wl_buffer);
+    }
 
     return EXIT_SUCCESS;
 }
